using Nemerle
using Nemerle.Utility
using Nemerle.Logging
using Nemerle.Collections
using Nemerle.Imperative
using Nemerle.Profiling

set namespace Fx7

public class Grinder
  pool : Term.Pool

  cache_neg : Hashtable [Term, Term] = Hashtable () 
  cache_pos : Hashtable [Term, Term] = Hashtable () 

  static CNFSize (f : Term, neg : bool) : int
    match (f.Name)
      | "and" when !neg \
      | "or" when neg =>
        mutable res = 0
        f.IterChildren (c => res += CNFSize (c, neg))
        res
      | "and" | "or" =>
        mutable res = 1
        f.IterChildren (c => res *= CNFSize (c, neg))
        res
      | "not" => CNFSize (f.OnlyChild, !neg)
      | _ => 1

  static LinSize (f : Term) : int
    match (f.Name)
      | "and" | "or" =>
        mutable res = 0
        f.IterChildren (c => res += LinSize (c))
        res
      | "not" => LinSize (f.OnlyChild)
      | _ => 1


  RemoveDuplicates(a : list [Term]) : list [Term]
    a.Sort(_.CompareTo(_)).RemoveDuplicates()

  MakeAndOr (head : string, children : list [Term]) : Term
    mutable seen_true = false
    mutable seen_false = false
    def children = children.Filter (t =>
      if (t === pool.PredFalse) { seen_false = true; false }
      else if (t === pool.PredTrue) { seen_true = true; false }
      else true
    )
    if (head == "and" && seen_false) pool.PredFalse
    else if (head == "or" && seen_true) pool.PredTrue
    else 
      match (children)
        | [] => if (head == "and") pool.PredTrue else pool.PredFalse
        | [one] => one
        | children => pool.Get (head, RemoveDuplicates (children))


  ContainsVar (var : Term, form : Term) : bool
    assert (var.Arity == 0)
    if (var === form) true
    else 
      when (form.Name == "forall")
        match (form.Children)
          | [vars, _] | [vars, _, _] =>
            when (vars.Children.Exists (_ == var : object))
              return false
          | _ => assert (false)
      form.Children.Exists (ContainsVar (var, _))

  BuildForall (vars : list [Term], pats : Term, body : Term) : Term
    def vars = pool.Get ("$@vars", vars)
    if (pats == null)
      pool.Get ("forall", [vars, body])
    else
      pool.Get ("forall", [vars, pats, body])

  GrindForall (vars : Term, pats : Term, body : Term) : Term
    assert (vars.Name == "$@vars")
    def (used, not_used) = vars.Children.Partition (ContainsVar (_, body))
    if (pats != null && not_used.Exists (ContainsVar (_, pats)))
      log (GRIND, $"the formula FORALL $vars. $body uses variables unused in body in the pattern: $pats")
      pool.Get ("forall", [vars, pats, body])
    else if (used is [])
      log (GRIND, $ "got rid of vars in FORALL $vars. $body")
      body
    else
      unless (not_used is [])
        log (GRIND, $ "droped $not_used from FORALL $vars. $body")
     
      def orig_vars = vars
      def grind_neq (pats, vars, clause)
        def is_var (v)
          v.Arity == 0 && vars.Contains (v)
        def is_ground (t)
          ! vars.Exists (ContainsVar (_, t))
        def is_neq (_ : Term)
          | E ("not", [E ("=", [t1, t2])]) =>
            if (t1 !== t2 && is_var (t1) && (is_var (t2) || is_ground (t2)))
              Some (t1, t2)
            else if (is_var (t2) && is_ground (t1))
              Some (t2, t1)
            else None ()
          | _ => None ()
        def find_neq (_)
          | (acc, t :: ts) =>
            match (is_neq (t))
              | Some as r => (r, acc.RevAppend (ts))
              | None => find_neq (t :: acc, ts)
          | (acc, []) => (None (), acc.Rev ())

        match (find_neq ([], clause))
          | (Some ((t1, t2)), rest) =>
            def subst = IntTree.Empty ().Add (t1.Id, t2)
            def rest = rest.Map (_.Apply (subst))
            def pats = if (pats == null) null else pats.Apply (subst)
            def vars = vars.Filter (_ != t1 : object)
            grind_neq (pats, vars, rest)
          | _ => 
            def body' = MakeAndOr ("or", clause)
            when (body !== body')
              log (GRIND, $"FORALL $orig_vars. $body ==> FORALL $vars. $(body')")
            if (vars is []) body'
            else BuildForall (vars, pats, body')
      
      if (Driver.GrindForall && body.Name == "or")
        grind_neq (pats, used, body.Children)
      else BuildForall (used, pats, body)
      
  EqualityPropagation (formula : Term) : Term
    def fixpoint_apply (form, subst)
      def form' = form.Apply (subst)
      //log (TEMP, $"subst: $form -> $(form')")
      if (form' === form) form
      else fixpoint_apply (form', subst)

    if (formula.Name == "and")
      def f (_, form, (non_eq, subst))
        if (form.Name == "=" && form.Child_1of2.Arity == 0 && form.Child_2of2.Arity == 0)
          //log (TEMP, $"form = $form")
          def (a, b) = (fixpoint_apply (form.Child_1of2, subst), 
                        fixpoint_apply (form.Child_2of2, subst))
          //log (TEMP, $" --> ($a, $b)")
          if (a === b) (non_eq, subst)
          else
            def (a, b) = 
              if (a.Depth < b.Depth) (b, a) 
              else if (a.Depth > b.Depth) (a, b)
              else if (a.Id < b.Id) (a, b) else (b, a)
            (non_eq, subst.Add (a.Id, b))
        else if (form.Arity == 0 && form.Name == "true") (non_eq, subst)
        else (form :: non_eq, subst)
      def (non_eq, subst) = formula.FoldIndexedChildren (([], IntTree.Empty ()), f)

      def expand (f, t, subst)
        def t' = fixpoint_apply (t, subst)
        if (t === t') subst
        else subst.Replace (f, t')
      def fixpoint (subst)
        def subst' = subst.Fold (subst, expand)
        if (subst' === subst) subst
        else fixpoint (subst')
      def subst = fixpoint (subst)

      def args = non_eq.RevMap (_.Apply (subst))
      //def args = args.Sort ((a, b) => a.Name.CompareTo (b.Name))
      MakeAndOr ("and", args)

    else formula

  SkolemizeForall (qvars : list [Term], sub : Subst, f : Term, neg : bool) : Term
    // log (TEMP, $ "skol: $qvars $neg $sub $f")
    match (f.Children)
      | [vars, pats, body] \
      | [vars, body] with pats = null =>
        if (neg == (f.Name == "forall"))
          def skolem_fn ()
            pool.Get (
              pool.FreshFn (qvars.Length),
              qvars)
          def sub = vars.Children.FoldLeft (sub, (v, sub) =>
             sub.Replace (v.Id, skolem_fn ()))
          //when (pats != null)
          //  log (WARN, $ "patterns given for q-forumula used in " 
          //               "existential position $f")
          Skolemize (qvars, sub, body, neg)
        else
          def qvars = 
            vars.Children.FoldLeft (qvars, (v, qvars) =>
               v :: qvars.Filter (_ != v : object))
          def pats =
            if (pats == null) null
            else
              Skolemize (qvars, sub, pats, neg)
          def body = Skolemize (qvars, sub, body, neg)

          def cnf_size = CNFSize (body, false)
          def lin_size = LinSize (body)
          def make_forall (body)
            GrindForall (vars, pats, body)

          if (cnf_size > 5 * lin_size)
            log (INFO, $ "not using CNF-conversion because of $cnf_size clauses required (size: $lin_size)")
            make_forall (body)
          else
            def cnf (f)
              match (f.Name)
                | "and" =>
                  $[ clause | ch in f.Children, clause in cnf (ch) ]

                | "or" =>
                  def loop (pref, sets)
                    match (sets)
                      | set :: sets =>
                        $[clause | c in set, clause in loop (c :: pref, sets)]
                      | [] => [List.Flatten (pref)]
                  loop ([], $[ cnf (ch) | ch in f.Children ])

                | _ => [[f]]

            def make (clause)
              def ht = Hashtable ()
              mutable is_true = false
              def check (l : Term)
                | E ("not", [idx]) \
                | idx =>
                  if (ht.Contains (idx))
                    unless (ht [idx] === l) is_true = true
                    false
                  else
                    ht [idx] = l
                    true
              def clause = clause.Filter (check)
              if (is_true)
                pool.PredTrue
              else if (clause is [])
                pool.PredFalse
              else
                make_forall (MakeAndOr ("or", clause))
            
            MakeAndOr ("and", cnf (body).Map (make))

      | _ => assert (false)

  Skolemize (qvars : list [Term], sub : Subst, f : Term, neg : bool) : Term
    def nnf (t)
      Skolemize (qvars, sub, t, neg)

    def cache = 
      if (qvars is [] && sub.IsEmpty)
        if (neg) cache_neg else cache_pos
      else null

    if (cache != null && cache.Contains (f))
      cache [f]
    else
      def res =
        if (sub.Contains (f.Id))
          sub.Get (f.Id)
        else
          match (f.Name)
            | "not" => Skolemize (qvars, sub, f.OnlyChild, !neg)

            | "iff" =>
              nnf (pool.Get ("and", [
                      pool.Get ("or", [pool.Get ("not", [f.Child_1of2]), f.Child_2of2]),
                      pool.Get ("or", [pool.Get ("not", [f.Child_2of2]), f.Child_1of2])]))

            | "implies" =>
              nnf (pool.Get ("or", [pool.Get ("not", [f.Child_1of2]), f.Child_2of2]))

            | "and" | "or" =>
              def head = if (neg) if (f.Name == "and") "or" else "and" else f.Name
              def get_children (_, c, acc)
                def c = nnf (c)
                if (c.Name == head)
                  c.Children + acc
                else c :: acc
              def children = f.FoldIndexedChildren ([], get_children)
              MakeAndOr (head, children)

            | "exists" | "forall" => 
              def res = SkolemizeForall (qvars, sub, f, neg)
              //log (TEMP, $"sf: $neg $f => $res")
              res
            | _ => 
              def f = f.Apply (sub)
              def f =
                match (f.Name)
                  | "=" | "<=" | ">=" =>
                    if (f.Child_1of2 === f.Child_2of2)
                      log (GRIND, $"simplifing $f to true")
                      pool.PredTrue
                    else f
                  | "<" | ">" =>
                    if (f.Child_1of2 === f.Child_2of2)
                      log (GRIND, $"simplifing $f to false")
                      pool.PredFalse
                    else f
                  | _ => f

              if (neg)
                if (f === pool.PredTrue)
                  pool.PredFalse
                else if (f === pool.PredFalse)
                  pool.PredTrue
                else
                  pool.Get ("not", [f])
              else f

      //log (TEMP, $"skol: $f -> $res")
      when (cache != null) cache [f] = res
      res
    
  
  public this (core : Core)
    pool = core.TermPool


  public Skolemize (f : Term) : Term
    if (Driver.DisableGrinder)
      f
    else
      def f = Skolemize ([], IntTree.Empty (), f, false)
      def f = if (Driver.EqProp) EqualityPropagation (f) else f
      // System.IO.File.WriteAllText ("eqprop-ok.smt", f.ToPrettyString ())
      f
  